@@ -8,11 +8,11 @@
 
 #include "dm-snapshot.h"
 #include "kcopyd.h"
+
 #include <linux/mm.h>
 #include <linux/pagemap.h>
-
-#define SECTOR_SIZE 512
-#define SECTOR_SHIFT 9
+#include <linux/vmalloc.h>
+#include <linux/slab.h>
 
 /*-----------------------------------------------------------------
  * Persistent snapshots, by persistent we mean that the snapshot
@@ -78,7 +78,7 @@ struct disk_exception {
 };
 
 struct commit_callback {
-	void (*callback)(void *, int success);
+	void (*callback) (void *, int success);
 	void *context;
 };
 
@@ -156,32 +156,6 @@ static int do_io(int rw, struct kcopyd_region *where, struct kiobuf *iobuf)
 	return 0;
 }
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION ( 2, 4, 19)
-/*
- * FIXME: Remove once 2.4.19 has been released.
- */
-struct page *vmalloc_to_page(void *vmalloc_addr)
-{
-	unsigned long addr = (unsigned long) vmalloc_addr;
-	struct page *page = NULL;
-	pmd_t *pmd;
-	pte_t *pte;
-	pgd_t *pgd;
-
-	pgd = pgd_offset_k(addr);
-	if (!pgd_none(*pgd)) {
-		pmd = pmd_offset(pgd, addr);
-		if (!pmd_none(*pmd)) {
-			pte = pte_offset(pmd, addr);
-			if (pte_present(*pte)) {
-				page = pte_page(*pte);
-			}
-		}
-	}
-	return page;
-}
-#endif
-
 static int allocate_iobuf(struct pstore *ps)
 {
 	size_t i, r = -ENOMEM, len, nr_pages;
@@ -472,10 +446,10 @@ static int persistent_prepare(struct exception_store *store,
 {
 	struct pstore *ps = get_info(store);
 	uint32_t stride;
-	offset_t size = get_dev_size(store->snap->cow->dev);
+	sector_t size = get_dev_size(store->snap->cow->dev);
 
 	/* Is there enough room ? */
-	if (size <= (ps->next_free * store->snap->chunk_size))
+	if (size < ((ps->next_free + 1) * store->snap->chunk_size))
 		return -ENOSPC;
 
 	e->new_chunk = ps->next_free;
@@ -485,7 +459,7 @@ static int persistent_prepare(struct exception_store *store,
 	 * into account the location of the metadata chunks.
 	 */
 	stride = (ps->exceptions_per_area + 1);
-	if (!(++ps->next_free % stride))
+	if ((++ps->next_free % stride) == 1)
 		ps->next_free++;
 
 	atomic_inc(&ps->pending_count);
@@ -616,6 +590,12 @@ int dm_create_persistent(struct exception_store *store, uint32_t chunk_size)
 		/*
 		 * Sanity checks.
 		 */
+		if (!ps->valid) {
+			DMWARN("snapshot is marked invalid");
+			r = -EINVAL;
+			goto bad;   
+		}
+
 		if (ps->chunk_size != chunk_size) {
 			DMWARN("chunk size for existing snapshot different "
 			       "from that requested");
@@ -664,7 +644,7 @@ int dm_create_persistent(struct exception_store *store, uint32_t chunk_size)
  * Implementation of the store for non-persistent snapshots.
  *---------------------------------------------------------------*/
 struct transient_c {
-	offset_t next_free;
+	sector_t next_free;
 };
 
 void transient_destroy(struct exception_store *store)
@@ -675,7 +655,7 @@ void transient_destroy(struct exception_store *store)
 int transient_prepare(struct exception_store *store, struct exception *e)
 {
 	struct transient_c *tc = (struct transient_c *) store->context;
-	offset_t size = get_dev_size(store->snap->cow->dev);
+	sector_t size = get_dev_size(store->snap->cow->dev);
 
 	if (size < (tc->next_free + store->snap->chunk_size))
 		return -1;
@@ -702,7 +682,7 @@ static int transient_percentfull(struct exception_store *store)
 }
 
 int dm_create_transient(struct exception_store *store,
-			struct dm_snapshot *s, int blocksize, void **error)
+			struct dm_snapshot *s, int blocksize)
 {
 	struct transient_c *tc;
 
