@@ -4,30 +4,26 @@
  * This file is released under the GPL.
  */
 
-#include <linux/config.h>
-#include <linux/module.h>
-#include <linux/init.h>
-#include <linux/slab.h>
-#include <linux/list.h>
-#include <linux/fs.h>
-#include <linux/blkdev.h>
-#include <linux/device-mapper.h>
-#include <linux/mempool.h>
 #include <asm/atomic.h>
-#include <linux/pagemap.h>
+
+#include <linux/blkdev.h>
+#include <linux/config.h>
+#include <linux/device-mapper.h>
+#include <linux/fs.h>
+#include <linux/init.h>
+#include <linux/list.h>
 #include <linux/locks.h>
+#include <linux/mempool.h>
+#include <linux/module.h>
+#include <linux/pagemap.h>
+#include <linux/slab.h>
+#include <linux/vmalloc.h>
 
 #include "kcopyd.h"
 
 /* FIXME: this is only needed for the DMERR macros */
 #include "dm.h"
 
-/*
- * Hard sector size used all over the kernel.
- */
-#define SECTOR_SIZE 512
-#define SECTOR_SHIFT 9
-
 static void wake_kcopyd(void);
 
 /*-----------------------------------------------------------------
@@ -63,8 +59,10 @@ static int init_pages(void)
 	return 0;
 
       bad:
-	while (i--)
+	while (i--) {
+		UnlockPage(_pages_array[i]);
 		__free_page(_pages_array[i]);
+	}
 	return -ENOMEM;
 }
 
@@ -255,7 +253,7 @@ struct kcopyd_job *kcopyd_alloc_job(void)
 {
 	struct kcopyd_job *job;
 
-	job = mempool_alloc(_job_pool, GFP_KERNEL);
+	job = mempool_alloc(_job_pool, GFP_NOIO);
 	if (!job)
 		return NULL;
 
@@ -332,7 +330,6 @@ static void dispatch_bh(struct kcopyd_job *job,
 	p = block >> job->bpp_shift;
 	block &= job->bpp_mask;
 
-	bh->b_dev = B_FREE;
 	bh->b_size = job->block_size;
 	set_bh_page(bh, job->pages[p], ((block << job->block_shift) +
 					job->offset) << SECTOR_SHIFT);
@@ -341,9 +338,11 @@ static void dispatch_bh(struct kcopyd_job *job,
 	init_buffer(bh, end_bh, job);
 
 	bh->b_dev = job->disk.dev;
-	bh->b_state = ((1 << BH_Mapped) | (1 << BH_Lock) | (1 << BH_Req));
+	atomic_set(&bh->b_count, 1);
+
+	bh->b_state = ((1 << BH_Uptodate) | (1 << BH_Mapped) |
+		       (1 << BH_Lock) | (1 << BH_Req));
 
-	set_bit(BH_Uptodate, &bh->b_state);
 	if (job->rw == WRITE)
 		clear_bit(BH_Dirty, &bh->b_state);
 
@@ -402,7 +401,7 @@ static int run_pages_job(struct kcopyd_job *job)
 	}
 
 	if (r == -ENOMEM)
-		/* can complete now */
+		/* can't complete now */
 		return 1;
 
 	return r;
@@ -611,7 +610,7 @@ static void exit_copier(void)
 
 static inline struct copy_info *alloc_copy_info(void)
 {
-	return mempool_alloc(_copy_pool, GFP_KERNEL);
+	return mempool_alloc(_copy_pool, GFP_NOIO);
 }
 
 static inline void free_copy_info(struct copy_info *info)
@@ -655,8 +654,10 @@ void copy_write(struct kcopyd_job *job)
 {
 	struct copy_info *info = (struct copy_info *) job->context;
 
-	if (job->err && info->notify) {
-		info->notify(job->err, job->context);
+	if (job->err) {
+		if (info->notify)
+			info->notify(job->err, job->context);
+
 		kcopyd_free_job(job);
 		free_copy_info(info);
 		return;
@@ -665,7 +666,6 @@ void copy_write(struct kcopyd_job *job)
 	job->rw = WRITE;
 	memcpy(&job->disk, &info->to, sizeof(job->disk));
 	job->callback = copy_complete;
-	job->context = info;
 
 	/*
 	 * Queue the write.
@@ -712,7 +712,6 @@ int kcopyd_write_pages(struct kcopyd_region *to, int nr_pages,
 
 	memcpy(&job->disk, &info->to, sizeof(job->disk));
 	job->offset = offset;
-	calc_block_sizes(job);
 	job->callback = page_write_complete;
 	job->context = info;
 
@@ -753,7 +752,6 @@ int kcopyd_copy(struct kcopyd_region *from, struct kcopyd_region *to,
 	memcpy(&job->disk, from, sizeof(*from));
 
 	job->offset = 0;
-	calc_block_sizes(job);
 	job->callback = copy_write;
 	job->context = info;
 
